/****************************************************************************
*
*    The MIT License (MIT)
*
*    Copyright (c) 2014 - 2022 Vivante Corporation
*
*    Permission is hereby granted, free of charge, to any person obtaining a
*    copy of this software and associated documentation files (the "Software"),
*    to deal in the Software without restriction, including without limitation
*    the rights to use, copy, modify, merge, publish, distribute, sublicense,
*    and/or sell copies of the Software, and to permit persons to whom the
*    Software is furnished to do so, subject to the following conditions:
*
*    The above copyright notice and this permission notice shall be included in
*    all copies or substantial portions of the Software.
*
*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
*    DEALINGS IN THE SOFTWARE.
*
*****************************************************************************
*
*    The GPL License (GPL)
*
*    Copyright (C) 2014 - 2022 Vivante Corporation
*
*    This program is free software; you can redistribute it and/or
*    modify it under the terms of the GNU General Public License
*    as published by the Free Software Foundation; either version 2
*    of the License, or (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software Foundation,
*    Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
*****************************************************************************
*
*    Note: This software is released under dual MIT and GPL licenses. A
*    recipient may use this file under the terms of either the MIT license or
*    GPL License. If you wish to use only one license not the other, you can
*    indicate your decision by deleting one of the above license notices in your
*    version of this file.
*
*****************************************************************************/


#include "gc_hal_kernel_linux.h"
#include "gc_hal_kernel_allocator.h"
#include <linux/pagemap.h>
#include <linux/seq_file.h>
#include <linux/mman.h>
#include <linux/atomic.h>
#include <linux/dma-mapping.h>
#include <linux/slab.h>
#include <linux/io.h>
#include <linux/ioport.h>

#include "gc_hal_kernel_linux.h"
#include "gc_hal_kernel_platform.h"
#include "gc_hal_kernel_platform_supernova.h"
#include <gc_hal_kernel_debug.h>
#include "common.h"

#if defined(USE_MEMORY_OSAL)
#include "memory_osal.h"
#else
#include "memory.h"

static int memory_task_id = 0;
#endif

#define _GC_OBJ_ZONE    gcvZONE_OS

#if !defined(USE_MEMORY_OSAL)
static struct memory_t *hdma_get_tmem_handler(gckALLOCATOR Allocator)
{
    struct _gcsPLATFORM_PCIE *pcie_platform;
    gcsPLATFORM * platform = Allocator->os->device->platform;
    struct sn_tranx_t *tdev;

    pcie_platform = (struct _gcsPLATFORM_PCIE *)platform;
    if (!pcie_platform)
        return NULL;

    tdev = (struct sn_tranx_t *)pci_get_drvdata(pcie_platform->pcie_info[0].pdev);
    if (!tdev)
        return NULL;

    return tdev->modules[SN_MODULE_MEMORY];
}

static void hdma_set_tmem_slice(gckALLOCATOR Allocator, int slice)
{
    struct _gcsPLATFORM_PCIE *pcie_platform;
    gcsPLATFORM * platform = Allocator->os->device->platform;
    struct sn_tranx_t *tdev;

    pcie_platform = (struct _gcsPLATFORM_PCIE *)platform;
    tdev = (struct sn_tranx_t *)pci_get_drvdata(pcie_platform->pcie_info[0].pdev);

    /* use default method if slice is wrong */
    if (slice < ALLOCATE_METHOD_S1 || slice > ALLOCATE_METHOD_S1S2)
        tdev->mem_alloc_method = sn_mem_get_method(tdev);
    else
        tdev->mem_alloc_method = slice;
}
#else
static struct sn_tranx_t* hdma_get_trans_dev(gckALLOCATOR Allocator)
{
    return (struct sn_tranx_t*) pci_get_drvdata(((struct _gcsPLATFORM_PCIE*) Allocator->os->device->platform)->pcie_info[0].pdev);
}
#endif

#ifdef CONFIG_DEBUG_FS
static int hdma_show(struct seq_file* m, void* data)
{
    gcsINFO_NODE *node = m->private;
    gckALLOCATOR Allocator = node->device;
    struct memory_t *tmem = gcvNULL;

    if (!Allocator)
        return -EINVAL;

    tmem = hdma_get_tmem_handler(Allocator);
    return 0;
}

static gcsINFO info_list[] =
{
    {"vsi-hdma", hdma_show},
};

#else
static ssize_t
hdma_usage_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    long long low = 0, high = 0;
    int len = 0;

    len = sprintf(buf + len, "type        n pages        bytes\n");
    len += sprintf(buf + len, "normal   %10llu %12llu\n", low, low * PAGE_SIZE);
    len += sprintf(buf + len, "HighMem  %10llu %12llu\n", high, high * PAGE_SIZE);

    return len;
}

static DEVICE_ATTR_RO(hdma_usage);

static struct attribute *vsi_hdma_attrs[] = {
    &dev_attr_hdma_usage.attr,
    NULL,
};

static struct attribute_group vsi_hdma_allocator_attribute_group = {
    .name = "vsi-hdma",
    .attrs = vsi_hdma_attrs,
};
#endif

static void
hdma_debugfs_init(
    IN gckALLOCATOR Allocator,
    IN gckDEBUGFS_DIR Root
    )
{
#ifdef CONFIG_DEBUG_FS
    gcmkVERIFY_OK(
        gckDEBUGFS_DIR_Init(&Allocator->debugfsDir, Root->root, "vsi-hdma"));

    gcmkVERIFY_OK(gckDEBUGFS_DIR_CreateFiles(
        &Allocator->debugfsDir,
        info_list,
        gcmCOUNTOF(info_list),
        Allocator
        ));
#else
    int ret;
    ret = sysfs_create_group(Allocator->os->allocatorSysfsDir, &vsi_hdma_allocator_attribute_group);
    if (ret < 0) {
        gcmkPRINT("failed to create sysfs for vsi-hdma.\n");
    }
#endif

}

static void
hdma_debugfs_cleanup(
    IN gckALLOCATOR Allocator
    )
{
#ifdef CONFIG_DEBUG_FS
    gcmkVERIFY_OK(gckDEBUGFS_DIR_RemoveFiles(
        &Allocator->debugfsDir,
        info_list,
        gcmCOUNTOF(info_list)
        ));

    gckDEBUGFS_DIR_Deinit(&Allocator->debugfsDir);
#else
    sysfs_remove_group(Allocator->os->allocatorSysfsDir, &vsi_hdma_allocator_attribute_group);
#endif
}

static gceSTATUS memory_alloc(
    IN gckALLOCATOR Allocator,
    unsigned long *busaddr,
    gctINT mdl_fd,
    size_t size
)
{
    int ret = 0;
#if !defined(USE_MEMORY_OSAL)
    struct memory_t *tmem = gcvNULL;
    static int last_alloc_slice = ALLOCATE_METHOD_S1;

    *busaddr = -1;
    tmem = hdma_get_tmem_handler(Allocator);
    if(!tmem) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_OS,"hdma: failed to get tmem\n");
        return  gcvSTATUS_OUT_OF_RESOURCES;
    }

    if (memory_task_id == 0) {
        memory_task_id = get_task_id(tmem, NULL);
        if (memory_task_id == -1) {
            gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_OS, "hdma: failed to get memory task id.\n");
            return gcvSTATUS_OUT_OF_RESOURCES;
        }
    }

    hdma_set_tmem_slice(Allocator, last_alloc_slice);
    if (last_alloc_slice == ALLOCATE_METHOD_S1)
        last_alloc_slice = ALLOCATE_METHOD_S2;
    else
        last_alloc_slice = ALLOCATE_METHOD_S1;

    ret = alloc_mem_ep(busaddr, size, memory_task_id, 0, tmem);

    /* recover default method */
    hdma_set_tmem_slice(Allocator, ALLOCATE_METHOD_S1S2 + 1);
#else
    ret = sn_mem_osal_alloc_internal(hdma_get_trans_dev(Allocator), busaddr, size);
#endif
    if (ret) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_OS, "hdma: alloc memory failed.\n");
        return -EFAULT;
    }

    if (*busaddr == -1) {
        gcmkTRACE_ZONE(
                       gcvLEVEL_INFO, gcvZONE_OS,
                       "hdma: Allocation FAILED: size = %ld\n",
                       size
                      );

        return  gcvSTATUS_OUT_OF_MEMORY;
    }
    return gcvSTATUS_OK;
}

/* Free a buffer based on bus address */
static int memory_free(
    IN gckALLOCATOR Allocator,
    unsigned long busaddr,
    gctINT task_id
    )
{
    int ret = 0;
#if !defined(USE_MEMORY_OSAL)
    struct memory_t *tmem = gcvNULL;

    tmem = hdma_get_tmem_handler(Allocator);
    if(!tmem) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_OS,"hdma: failed to get tmem\n");
        return gcvSTATUS_OUT_OF_RESOURCES;
    }

    ret = free_mem_ep(busaddr, 0, memory_task_id, tmem);
#else
    ret = sn_mem_osal_free_internal(hdma_get_trans_dev(Allocator), busaddr);
#endif
    if (ret) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_OS, "hdma: free memory failed\n");
        return -EFAULT;
    }
    return 0;
}

static gceSTATUS
hdma_alloc(
    IN gckALLOCATOR Allocator,
    INOUT PLINUX_MDL Mdl,
    IN gctSIZE_T NumPages,
    IN gctUINT32 Flags
    )
{
    gceSTATUS status;
    unsigned long bus_addr;
    gctSIZE_T bytes;

    gcmkHEADER_ARG("Allocator=%p Mdl=%p NumPages=%zu Flags=0x%x", Allocator, Mdl, NumPages, Flags);

    bytes = (gctSIZE_T)(NumPages << PAGE_SHIFT);
    status = memory_alloc(Allocator, &bus_addr, Mdl->mdl_fd, bytes);
    if(status != gcvSTATUS_OK)
    {
        gcmkTRACE_ZONE(
                       gcvLEVEL_INFO, gcvZONE_OS,
                       "hdma: memory alloc failed\n"
                      );
        goto OnError;
    }

    Mdl->numPages = NumPages;
    Mdl->phy_addr = bus_addr;

    gcmkTRACE_ZONE(
        gcvLEVEL_INFO, gcvZONE_OS,
        "hdma: alloc = 0x%llx, size = 0x%lx\n",
        Mdl->phy_addr, (Mdl->numPages) << PAGE_SHIFT
        );

    gcmkPRINT("\t\t hdma: alloc = 0x%llx, size = 0x%lx\n\n", Mdl->phy_addr, (Mdl->numPages) << PAGE_SHIFT);

    status = gcvSTATUS_OK;

OnError:
    gcmkFOOTER();
    return status;
}

static gceSTATUS
hdma_attach(
    IN gckALLOCATOR Allocator,
    IN gcsATTACH_DESC_PTR Desc,
    IN PLINUX_MDL Mdl
    )
{
    gceSTATUS status;

    gcmkHEADER_ARG("Allocator=%p Desc=%p Mdl=%p", Allocator, Desc, Mdl);
    if (Desc == gcvNULL)
    {
        status = gcvSTATUS_INVALID_ARGUMENT;
        goto OnError;
    }


    status = gcvSTATUS_OK;

OnError:
    gcmkFOOTER();
    return status;
}

static void
hdma_free(
    IN gckALLOCATOR Allocator,
    IN OUT PLINUX_MDL Mdl
    )
{
    gceSTATUS status;

    gcmkPRINT("hdma memory free 0x%lx\n", (unsigned long)(Mdl->phy_addr));
    status = memory_free(Allocator, (unsigned long)(Mdl->phy_addr), Mdl->mdl_fd);
    if (status) {
        gcmkPRINT("hdma memory free failed(%d)\n", status);
    }
}


static gceSTATUS
hdma_mmap(
    IN gckALLOCATOR Allocator,
    IN PLINUX_MDL Mdl,
    IN gctBOOL Cacheable,
    IN gctSIZE_T skipPages,
    IN gctSIZE_T numPages,
    IN struct vm_area_struct *vma
    )
{
    gceSTATUS status = gcvSTATUS_OK;

    gcmkHEADER_ARG("Allocator=%p Mdl=%p vma=%p", Allocator, Mdl, vma);
    gcmkTRACE(gcvLEVEL_ERROR,"memory on device side can't be mapped and accessed by host PC\n");

    status = gcvSTATUS_OUT_OF_MEMORY;

    gcmkFOOTER();
    return status;
}

static void
hdma_unmap_user(
    IN gckALLOCATOR Allocator,
    IN PLINUX_MDL Mdl,
    IN PLINUX_MDL_MAP MdlMap,
    IN gctUINT32 Size
    )
{

    if (unlikely(!current->mm))
        return;

    if (vm_munmap((unsigned long)MdlMap->vmaAddr, Mdl->numPages <<PAGE_SHIFT) < 0)
    {
    }
}

static gceSTATUS
hdma_map_user(
    gckALLOCATOR Allocator,
    PLINUX_MDL Mdl,
    PLINUX_MDL_MAP MdlMap,
    gctBOOL Cacheable
    )
{
    gctPOINTER userLogical = gcvNULL;
    gceSTATUS status = gcvSTATUS_OK;

    gcmkHEADER_ARG("Allocator=%p Mdl=%p Cacheable=%d", Allocator, Mdl, Cacheable);

    userLogical = (gctPOINTER)vm_mmap(NULL, 0L, (unsigned long)Mdl->phy_addr,
                PROT_READ | PROT_WRITE, MAP_SHARED | MAP_NORESERVE, 0);

    gcmkTRACE_ZONE(
        gcvLEVEL_INFO, gcvZONE_OS,
        "%s(%d): vmaAddr->%p for phys_addr->%p",
        __FUNCTION__, __LINE__, userLogical, Mdl
        );

    if (IS_ERR(userLogical))
    {
        gcmkTRACE_ZONE(
            gcvLEVEL_INFO, gcvZONE_OS,
            "%s(%d): do_mmap_pgoff error",
            __FUNCTION__, __LINE__
            );

        gcmkONERROR(gcvSTATUS_OUT_OF_MEMORY);
    }

    down_write(&current_mm_mmap_sem);
    do
    {
        struct vm_area_struct *vma = find_vma(current->mm, (unsigned long)userLogical);

        if (vma == gcvNULL)
        {
            gcmkTRACE_ZONE(
                gcvLEVEL_INFO, gcvZONE_OS,
                "%s(%d): find_vma error",
                __FUNCTION__, __LINE__
                );

            gcmkERR_BREAK(gcvSTATUS_OUT_OF_RESOURCES);
        }

        gcmkERR_BREAK(hdma_mmap(Allocator, Mdl, gcvFALSE, 0, Mdl->numPages, vma));

        MdlMap->vmaAddr = userLogical;
        MdlMap->cacheable = gcvFALSE;
        MdlMap->vma = vma;
    }
    while (gcvFALSE);
    up_write(&current_mm_mmap_sem);

OnError:
    if (gcmIS_ERROR(status) && userLogical)
    {
        hdma_unmap_user(Allocator, Mdl, userLogical, Mdl->numPages << PAGE_SHIFT);
    }
    gcmkFOOTER();
    return status;
}

static gceSTATUS
hdma_map_kernel(
    IN gckALLOCATOR Allocator,
    IN PLINUX_MDL Mdl,
    IN gctSIZE_T Offset,
    IN gctSIZE_T Bytes,
    OUT gctPOINTER *Logical
    )
{
    void *vaddr;

    vaddr = memremap(Mdl->phy_addr + Offset, Bytes, MEMREMAP_WC);

    if (!vaddr)
    {
        return gcvSTATUS_OUT_OF_MEMORY;
    }

    *Logical = vaddr;
    return gcvSTATUS_OK;
}

static gceSTATUS
hdma_unmap_kernel(
    IN gckALLOCATOR Allocator,
    IN PLINUX_MDL Mdl,
    IN gctPOINTER Logical
    )
{
    memunmap((void *)Logical);

    memory_free(Allocator,(unsigned long)(Mdl->phy_addr), Mdl->mdl_fd);
    return gcvSTATUS_OK;
}

static gceSTATUS
hdma_cache_op(
    IN gckALLOCATOR Allocator,
    IN PLINUX_MDL Mdl,
    IN gctSIZE_T Offset,
    IN gctPOINTER Logical,
    IN gctSIZE_T Bytes,
    IN gceCACHEOPERATION Operation
    )
{
    /* Always WC or UC, safe to use mb. */
    switch (Operation)
    {
    case gcvCACHE_CLEAN:
    case gcvCACHE_FLUSH:
        _MemoryBarrier();
        break;
    case gcvCACHE_INVALIDATE:
        break;
    default:
        return gcvSTATUS_INVALID_ARGUMENT;
    }

    return gcvSTATUS_OK;
}

static gceSTATUS
hdma_get_physical(
    IN gckALLOCATOR Allocator,
    IN PLINUX_MDL Mdl,
    IN gctUINT32 Offset,
    OUT gctPHYS_ADDR_T * Physical
    )
{
    *Physical = Mdl->phy_addr + Offset;
    return gcvSTATUS_OK;
}

static void
hdma_dtor(
    gcsALLOCATOR *Allocator
    )
{
    hdma_debugfs_cleanup(Allocator);
    if (Allocator->privateData)
        kfree(Allocator->privateData);

    kfree(Allocator);
}

/* EDMA allocator operations. */
static gcsALLOCATOR_OPERATIONS hdma_ops = {
    .Alloc              = hdma_alloc,
    .Free               = hdma_free,
    .Attach             = hdma_attach,
    .Mmap               = hdma_mmap,
    .MapUser            = hdma_map_user,
    .UnmapUser          = hdma_unmap_user,
    .MapKernel          = hdma_map_kernel,
    .UnmapKernel        = hdma_unmap_kernel,
    .Cache              = hdma_cache_op,
    .Physical           = hdma_get_physical,
};


/* GFP allocator entry. */
gceSTATUS
_VsiHdmaAllocatorInit(
    IN gckOS Os,
    IN gcsDEBUGFS_DIR *Parent,
    OUT gckALLOCATOR * Allocator
    )
{
    gceSTATUS status;
    gckALLOCATOR allocator = gcvNULL;

    gcmkONERROR(
        gckALLOCATOR_Construct(Os, &hdma_ops, &allocator));

    /* Register private data. */
    allocator->destructor = hdma_dtor;

    hdma_debugfs_init(allocator, Parent);

    allocator->capability = gcvALLOC_FLAG_HDMA;

#if defined(gcdEMULATE_SECURE_ALLOCATOR)
    allocator->capability |= gcvALLOC_FLAG_SECURITY;
#endif

    *Allocator = allocator;

    return gcvSTATUS_OK;

OnError:
    if (allocator)
        kfree(allocator);

    return status;
}
