// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2020 Xilinx Inc.
 *
 * This is xilinx scaler driver for Linux.
 * This file provide register operation and initialization,
 * like read/write a register or pull/push a batch of registers.
 */

#include <linux/pci.h>
#include <linux/pagemap.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/version.h>

#include "common.h"
#include "transcoder.h"
#include "regs.h"

#include "sn_perf.h"

#if defined(USE_OSAL)
#include "sn_osal.h"
#endif

//------------------------Address Info-------------------
// 0x00 : Control signals
//        bit 0  - ap_start (Read/Write/COH)
//        bit 1  - ap_done (Read/COR)
//        bit 2  - ap_idle (Read)
//        bit 3  - ap_ready (Read)
//        bit 7  - auto_restart (Read/Write)
//        others - reserved
// 0x04 : Global Interrupt Enable Register
//        bit 0  - Global Interrupt Enable (Read/Write)
//        others - reserved
// 0x08 : IP Interrupt Enable Register (Read/Write)
//        bit 0  - Channel 0 (ap_done)
//        bit 1  - Channel 1 (ap_ready)
//        others - reserved
// 0x0c : IP Interrupt Status Register (Read/TOW)
//        bit 0  - Channel 0 (ap_done)
//        bit 1  - Channel 1 (ap_ready)
//        others - reserved
// 0x10 : Feature select
//        bit 0  - crc en_dis
//        bit 1  - wdt en_dis
//        others - wdt timer threshold value in cycles
// 0x14 : reserved
// 0x18 : Data signal of cmd_blk_V
//        bit 31~0 - cmd_blk_V[31:0] (Read/Write)
// 0x1c : Data signal of cmd_blk_V
//        bit 31~0 - cmd_blk_V[63:32] (Read/Write)
// 0x8c : reserved
// 0x90 : Data signal of error_flag_crc_V
//        bit 0  - error_flag_crc_V[0] (Read)
//        others - reserved
// 0x94 : Control signal of error_flag_crc_V
//        bit 0  - error_flag_crc_V_ap_vld (Read/COR)
//        others - reserved
// 0x98 : Data signal of error_flag_wdt_V
//        bit 0  - error_flag_wdt_V[0] (Read)
//        others - reserved
// 0x9c : Control signal of error_flag_wdt_V
//        bit 0  - error_flag_wdt_V_ap_vld (Read/COR)
//        others - reserved
// 0xb0 : Version Register
//        bit 31~0 - ABRScaler IP Version (Read)
// 0xb4 : Frame Cycle Count_0
//        bit 31~0 - cycle count (Read)
//.
//.
// 0xf4 : Frame Cycle Count_15
//        bit 31~0 - cycle count (Read)

#define XABR_SC_ADDR_AP_CTRL            0x00
#define XABR_SC_ADDR_GIE                0x04
#define XABR_SC_ADDR_IER                0x08
#define XABR_SC_ADDR_ISR                0x0c
#define XABR_SC_ADDR_CRC_WDT            0x10
#define XABR_SC_ADDR_CMD_BLK_ADDR       0x18
#define XABR_SC_ADDR_DATA_CRC_ERROR     0x90
#define XABR_SC_ADDR_CTL_CRC_ERROR      0x94
#define XABR_SC_ADDR_DATA_WDT_ERROR     0x98
#define XABR_SC_ADDR_CTL_WDT_ERROR      0x9c
#define XABR_SC_ADDR_VERSION            0xb0
#define XABR_SC_ADDR_CYCLE_CNT_0        0xb4
#define XABR_SC_ADDR_CYCLE_CNT_15       0xf4
#define XABR_SC_REGMAP_SIZE             0xf8

#define MAX_CHANNELS                    16

#define XABR_CORE_IO_SIZE               (XABR_SC_REGMAP_SIZE)
#define XABR_DEC400_IO_SIZE             (1568*4)

typedef enum {
    XABR_CORE_HW = 0,
    XABR_CORE_DEC400,
    XABR_CORE_MAX
}xabr_core_type;

typedef enum {
    wd_ctl = 0,
    wd_data,
    crc_ctl,
    crc_data,
    ap_done,
    NUM_IRQ_FLAGS
}IrqFlags;

typedef enum {
    IRQ_SRC_WDT = 1,
    IRQ_SRC_CRC,
    IRQ_SRC_AP_DONE
}IrqSource;

struct xabr_core_desc {
    int slice;
    xabr_core_type type;
    long base;
    int iosize;
    int irq;
    const char *name;
};

/*config cores in xabr subsystem in 2 slices*/
//PF Mode
static struct xabr_core_desc pf_core_array[][XABR_CORE_MAX]= {
    {
        {0, XABR_CORE_HW,     S1_ABR_SCL_CORE_OFF,   XABR_CORE_IO_SIZE,   17, "xabr_scl_0"},
        {0, XABR_CORE_DEC400, S1_ABR_SCL_DEC400_OFF, XABR_DEC400_IO_SIZE, -1, "xabr_dec_0"},
    },
    {
        {1, XABR_CORE_HW,     S2_ABR_SCL_CORE_OFF,   XABR_CORE_IO_SIZE,   27, "xabr_scl_1"},
        {1, XABR_CORE_DEC400, S2_ABR_SCL_DEC400_OFF, XABR_DEC400_IO_SIZE, -1, "xabr_dec_1"},
    }
};

//One VF Mode
static struct xabr_core_desc onevf_core_array[][XABR_CORE_MAX]= {
    {
        {0, XABR_CORE_HW,     ONE_VF_S1_ABR_SCL_CORE_OFF,   XABR_CORE_IO_SIZE,   17, "xabr_scl_0"},
        {0, XABR_CORE_DEC400, ONE_VF_S1_ABR_SCL_DEC400_OFF, XABR_DEC400_IO_SIZE, -1, "xabr_dec_0"},
    },
    {
        {1, XABR_CORE_HW,     ONE_VF_S2_ABR_SCL_CORE_OFF,   XABR_CORE_IO_SIZE,   27, "xabr_scl_1"},
        {1, XABR_CORE_DEC400, ONE_VF_S2_ABR_SCL_DEC400_OFF, XABR_DEC400_IO_SIZE, -1, "xabr_dec_1"},
    }
};

struct core_cfg {
    int irq;
    unsigned long base_addr;
    volatile u8 *hwreg;
    u32 iosize;
    u32 *shadow;
    const char *name;
    int irq_recvd;
};

struct xabr_scaler {
    int num_cores;
    struct core_cfg subsys[MAX_SLICE_NUM][XABR_CORE_MAX];
    spinlock_t irq_lock;
    spinlock_t reserve_lock;
    wait_queue_head_t wait_queue;
    wait_queue_head_t reserve_queue;
    void *regs_shadow;
    IrqFlags irq_flags[MAX_SLICE_NUM][NUM_IRQ_FLAGS];
    int isCoreIdle[MAX_SLICE_NUM];
    int isCoreLocked[MAX_SLICE_NUM];
    struct file *filp[MAX_SLICE_NUM];
    u32 cycle_cnt[MAX_SLICE_NUM][MAX_CHANNELS];
    struct sn_tranx_t *tdev;
	struct loading_info loading[MAX_SLICE_NUM];
	struct timer_list loading_timer;
    SnPerfHandle perf_handle;
    int perf_enable;
};

static struct xabr_core_desc* get_core_desc(struct xabr_scaler *tabr, int *num_cores)
{
    struct xabr_core_desc *cp = NULL;
    int ncores = 0;

    switch(tabr->tdev->pf_vf_mode)
    {
        case PF_MODE:
            if (tabr->tdev->vf_index != PF_INDEX) {
                sn_pri(tabr->tdev, SN_ERR, "(%s)vf_index incorrect in PF_MODE\n", __func__);
            } else {
                ncores = sizeof(pf_core_array)/sizeof(pf_core_array[0]);
                cp     = &pf_core_array[0][0];
            }
            break;

        case ONE_VF_MODE:
            if (tabr->tdev->vf_index != VF1_INDEX) {
                sn_pri(tabr->tdev, SN_ERR, "(%s)vf_index incorrect in ONE_VF_MODE\n", __func__);
            } else {
                ncores = sizeof(onevf_core_array)/sizeof(onevf_core_array[0]);
                cp     = &onevf_core_array[0][0];
            }
            break;

        default:
            sn_pri(tabr->tdev, SN_ERR, "(%s)incorrect pf_vf_mode (%d)\n", __func__, tabr->tdev->pf_vf_mode);
            break;
    }

    *num_cores = ncores;
    return cp;
}


static int init_core_array(struct xabr_scaler *tabr)
{
    int i, j, irq;
    u32 total_io_size = 0;
    struct xabr_core_desc *cd = NULL;
    int num_cores = 0;

    //get core description table for pf/vf
    cd = get_core_desc(tabr, &num_cores);
    if (!num_cores) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: failed to get num_cores\n");
        return 0;
    }

    tabr->num_cores = num_cores;
    for (i = 0; i < MAX_SLICE_NUM; ++i) {
        for (j = 0; j < tabr->num_cores; j++) {

            tabr->subsys[i][j].base_addr = (cd+i*XABR_CORE_MAX+j)->base;
            tabr->subsys[i][j].iosize    = (cd+i*XABR_CORE_MAX+j)->iosize;
            irq                          = (cd+i*XABR_CORE_MAX+j)->irq;
            if (irq != -1)
                tabr->subsys[i][j].irq   = pci_irq_vector(tabr->tdev->pdev, irq);

            total_io_size               += tabr->subsys[i][j].iosize;
            tabr->subsys[i][j].hwreg     = 0;
            tabr->subsys[i][j].name      = (cd+i*XABR_CORE_MAX+j)->name;

            sn_pri(tabr->tdev, SN_DBG, "xabr:slice[%d][%d]: base=%x  iosize=%5d  tot_size=%5d  core_irq id: %4d  pci_irq: %4d\n",
                   i, j, tabr->subsys[i][j].base_addr, tabr->subsys[i][j].iosize, total_io_size,
                   irq, tabr->subsys[i][j].irq);
        }
    }

    return total_io_size;
}

static int reserve_io(struct xabr_scaler *tabr)
{
    int i, j;

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        for (j = 0; j < tabr->num_cores; j++) {
            if (tabr->subsys[i][j].iosize) {
                sn_pri(tabr->tdev, SN_DBG, "xabr:slice[%d][%d]: bar2_virt=0x%x  ipbase=0x%lx, iosize=%d\n", i, j,
                       tabr->tdev->bar2_virt, tabr->subsys[i][j].base_addr, tabr->subsys[i][j].iosize);

                tabr->subsys[i][j].hwreg = tabr->tdev->bar2_virt + tabr->subsys[i][j].base_addr;

                if (tabr->subsys[i][j].hwreg == NULL) {
                    sn_pri(tabr->tdev, SN_ERR, "xabr:slice[%d][%d]: failed to ioremap HW %d regs\n", i,j,j);
                    return -EBUSY;
                }
                sn_pri(tabr->tdev, SN_DBG, "xabr:slice[%d][%d]: hwreg(bar2_virt+base)=0x%x  reg[0]=0x%08x\n",
                       i, j, tabr->subsys[i][j].hwreg, readl(tabr->subsys[i][j].hwreg));
            }
        }
    }
    return 0;
}

static long xabr_write_regs(struct xabr_scaler *tabr, struct core_desc *core)
{
    long ret;
    u32 i, reg_cnt;
    u32 idx  = core->id;
    u32 type = core->type;
    struct reg_desc *regs_info;

    if (idx  >= MAX_SLICE_NUM              ||
        type >= XABR_CORE_MAX              ||
        !tabr->subsys[idx][type].base_addr ||
        !tabr->subsys[idx][type].hwreg) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: %s failed core_id: [%d][%d] base_addr: %uld\n",
               __func__, idx,type,tabr->subsys[idx][type].base_addr);
        return -EINVAL;
    }

    regs_info = (struct reg_desc *)tabr->subsys[idx][type].shadow;

    ret = copy_from_user(regs_info, core->regs, core->size);
    if (ret) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: %s failed %ld\n", __func__, ret);
        return -EFAULT;
    }

    reg_cnt = core->size / sizeof(struct reg_desc);
    for (i = 0; i < reg_cnt; i++) {
        writel(regs_info[i].val,  tabr->subsys[idx][type].hwreg + regs_info[i].id);

        if(type == XABR_CORE_HW) {
            //update core status, if frame start requested
            if((regs_info[i].id == XABR_SC_ADDR_AP_CTRL) &&
               ((regs_info[i].val & 0x01) == 0x01)) {
                spin_lock(&tabr->reserve_lock);
                tabr->isCoreIdle[idx] = 0;
                spin_unlock(&tabr->reserve_lock);
            }
        }
    }
    return 0;
}

static long xabr_read_regs(struct xabr_scaler *tabr, struct core_desc *core)
{
    long ret;
    u32 i, reg_cnt;
    u32 idx  = core->id;
    u32 type = core->type;
    struct reg_desc *regs_info;

    if (idx  >= MAX_SLICE_NUM              ||
        type >= XABR_CORE_MAX              ||
        !tabr->subsys[idx][type].base_addr ||
        !tabr->subsys[idx][type].hwreg) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: %s failed core_id: [%d][%d] base_addr: %uld\n",
               __func__, idx,type,tabr->subsys[idx][type].base_addr);
        return -EINVAL;
    }

    regs_info = (struct reg_desc *)tabr->subsys[idx][type].shadow;

    ret = copy_from_user(regs_info, core->regs, core->size);
    if (ret) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: %s failed %ld\n", __func__, ret);
        return -EFAULT;
    }

    reg_cnt = core->size / sizeof(struct reg_desc);
    /* read specific registers from hardware */
    for (i = 0; i < reg_cnt; i++) {
        regs_info[i].val = readl(tabr->subsys[idx][type].hwreg + regs_info[i].id);
    }

    /* put registers to user space*/
    ret = copy_to_user(core->regs, regs_info, core->size);
    if (ret) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: %s failed %ld\n", __func__, ret);
        return -EFAULT;
    }

    return 0;
}


static int check_xabr_irq(struct xabr_scaler *tabr, u32 slice, u32 *irq_status)
{
    unsigned long flags;
    volatile u8 *hwregs;
    int i,rdy = 0;

    spin_lock_irqsave(&tabr->irq_lock, flags);

    if (tabr->subsys[slice][XABR_CORE_HW].irq_recvd) {
        /* reset the wait condition(s) */
        tabr->subsys[slice][XABR_CORE_HW].irq_recvd = 0;
        rdy = 1;

        hwregs = tabr->subsys[slice][XABR_CORE_HW].hwreg;
        /* Decode IRQ Source
            - check if irq was WDT error
            - check if irq was CRC error
            - check if irq was frame done
        */
        if((tabr->irq_flags[slice][wd_ctl]  & 0x01) && (tabr->irq_flags[slice][wd_data] & 0x01)) {
            irq_status[0] = IRQ_SRC_WDT;
        } else if((tabr->irq_flags[slice][crc_ctl]  & 0x01) && (tabr->irq_flags[slice][crc_data] & 0x01)) {
            irq_status[0] = IRQ_SRC_CRC;
        } else if(tabr->irq_flags[slice][ap_done] & 0x02) {
            irq_status[0] = IRQ_SRC_AP_DONE;
        } else {
            irq_status[0] = 0;
        }

        //add cycle count
        for (i=0; i<MAX_CHANNELS; ++i) {
            irq_status[i+1] = tabr->cycle_cnt[slice][i];
        }

        //clear shadow irq flags
        for (i=0; i<NUM_IRQ_FLAGS; ++i) {
            tabr->irq_flags[slice][i] = 0;
        }

        //unmask interrupt for slice
        writel(0x01, hwregs + XABR_SC_ADDR_GIE);
        wake_up_interruptible_all(&tabr->reserve_queue);
    }

    spin_unlock_irqrestore(&tabr->irq_lock, flags);
    return rdy;
}

static long wait_xabr_ready(struct xabr_scaler *tabr,
                            u32 slice,
                            u32 timeout_sec,
                            u32 *irq_status)
{
    long ret;

    ret = wait_event_interruptible_timeout(tabr->wait_queue,
                                           check_xabr_irq(tabr, slice, irq_status),
                                           timeout_sec*HZ);

    if(!ret) {
        //timeout elapsed
        sn_pri(tabr->tdev, SN_ERR, "xabr: wait_for_done timed out\n");
    } else if (ret > 0) {
        ret = 0; //core ready
    } else { //-ERESTARTSYS
        sn_pri(tabr->tdev, SN_DBG,
            "xabr: wait_event_interruptible interrupted\n", ret);
    }
    return ret;
}


static irqreturn_t xabr_dec400_isr(int irq, void *dev_id)
{
    struct xabr_scaler *tabr = (struct xabr_scaler *) dev_id;

    sn_pri(tabr->tdev, SN_ERR,
            "xabr: xabr dec400 irq not supported yet\n");
    return IRQ_RETVAL(0);
}


static irqreturn_t xabr_scaler_isr(int irq, void *dev_id)
{
    unsigned long flags;
    unsigned int handled = 0;
    int j,slice = 0;
    volatile u8 *hwregs = NULL;
    int regid, idx;
    u32 int_shadow_isr_off = S1_INT_SD_XABR_ISR;
    u32 reg_read = 0;

    struct xabr_scaler *tabr = (struct xabr_scaler *) dev_id;

    spin_lock_irqsave(&tabr->irq_lock, flags);

    //detrmine slice index from irq#
    if(irq == tabr->subsys[0][XABR_CORE_HW].irq) {
        slice = 0;
        int_shadow_isr_off = S1_INT_SD_XABR_ISR;
    } else if(irq == tabr->subsys[1][XABR_CORE_HW].irq) {
        slice = 1;
        int_shadow_isr_off = S2_INT_SD_XABR_ISR;
    } else {
        sn_pri(tabr->tdev, SN_ERR, "xabr: error:: irq %d not supported \n", irq);
        goto end;
    }
    hwregs = tabr->subsys[slice][XABR_CORE_HW].hwreg;

    //zsp generates spurious interrupts for slice 1 that needs to be filtered here
    reg_read = readl(hwregs + XABR_SC_ADDR_ISR);
    if (reg_read != 1) {
        //sn_pri(tabr->tdev, SN_INF, "xabr: warn:: unknown IRQ_(%d)_src %d recvd on slice %d \n", irq, reg_read, slice);
        goto end;
    }
    //only ap_done is the valid irq source

    //read irq source registers into shadow regs
    tabr->irq_flags[slice][wd_ctl]   = readl(hwregs + XABR_SC_ADDR_CTL_WDT_ERROR);
    tabr->irq_flags[slice][wd_data]  = readl(hwregs + XABR_SC_ADDR_DATA_WDT_ERROR);
    tabr->irq_flags[slice][crc_ctl]  = readl(hwregs + XABR_SC_ADDR_CTL_CRC_ERROR);
    tabr->irq_flags[slice][crc_data] = readl(hwregs + XABR_SC_ADDR_DATA_CRC_ERROR);
    tabr->irq_flags[slice][ap_done]  = readl(hwregs + XABR_SC_ADDR_AP_CTRL);

    //read performance counters (skip unused registers in middle)
    idx = 0;
    for (j=0; j<MAX_CHANNELS+1; ++j) {
        regid = (XABR_SC_ADDR_CYCLE_CNT_0+j*4);
        if(regid == 0xcc) {//check for unused register in middle
            continue;
        }
        tabr->cycle_cnt[slice][idx]  = readl(hwregs + regid);
        ++idx;
    }
    handled++;
    tabr->subsys[slice][XABR_CORE_HW].irq_recvd = 1;
    //Mask interrupt line for the slice
    writel(0x00, hwregs + XABR_SC_ADDR_GIE);
    //if done_irq clear the ISR
    if (tabr->irq_flags[slice][ap_done] & 0x02) {
        writel(0x01, hwregs + XABR_SC_ADDR_ISR);
    }

    //irq decoding done in bottom half
end:
    spin_unlock_irqrestore(&tabr->irq_lock, flags);
    if(handled)
        wake_up_interruptible_all(&tabr->wait_queue);

    return IRQ_RETVAL(handled);
}

static void record_perf_event(struct xabr_scaler *tabr, u32 slice, SN_PERF_TYPE type)
{
    sn_perf_event event;
    event.slice = slice;
    event.ipId = SN_PERF_ID_XABR;
    event.unitId = 0;
    event.type = type;
    event.hwTsValid = 0;
    event.name = 0;
    sn_perf_record(tabr->perf_handle, &event);
}

static int get_core_idle(struct xabr_scaler *tabr, struct file *filp, u32 core_id, int *is_available)
{
    int ret = 0;
    int slice_mask = core_id>>8 & 0xFF; //0: slice_0, 1: slice_1: 2: Any
    int num_slices = ((slice_mask > 1) ? 2 : 1);
    int slice_idx  = ((num_slices > 1) ? 0 : slice_mask);
    int i;

    spin_lock(&tabr->reserve_lock);
    for (i=0; i<num_slices; ++i) {
        if (!tabr->isCoreLocked[slice_idx] && tabr->isCoreIdle[slice_idx]) {
            *is_available                 = slice_idx;
            if (tabr->perf_enable) {
                record_perf_event(tabr, slice_idx, SN_PERF_TYPE_START);
            }
            tabr->isCoreLocked[slice_idx] = 1;
            tabr->filp[slice_idx] = filp;
            ret = 1;
            break;
        }
        ++slice_idx;
    }
    spin_unlock(&tabr->reserve_lock);
    return ret;
}

static long xabr_reserve_core(struct xabr_scaler *tabr,
                              struct file *filp,
                              u32  core_id,
                              u32  timeout_sec,
                              int  *lock_status)
{
    long ret;

    /* lock core if available */
    ret = wait_event_interruptible_timeout(tabr->reserve_queue,
                                           get_core_idle(tabr, filp, core_id, lock_status),
                                           timeout_sec*HZ);
    if(!ret) {
        //timeout elapsed
        sn_pri(tabr->tdev, SN_ERR, "xabr: reserve_core %d timed out\n", core_id);
    } else if (ret > 0) {
        ret = 0; //core ready
    } else { //-ERESTARTSYS
        sn_pri(tabr->tdev, SN_DBG,
            "xabr: wait_event_interruptible interrupted\n");
    }

	if (core_id == 0)
		tabr->loading[0].tv_s = ktime_get();
	else
		tabr->loading[1].tv_s = ktime_get();
    return ret;
}

static long xabr_release_core(struct xabr_scaler *tabr, u32 slice)
{
    if (slice >= MAX_SLICE_NUM) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: %s invalid slice id %d\n",__func__, slice);
        return -EINVAL;
    }

    //update core status to idle and clear lock state
    spin_lock(&tabr->reserve_lock);
    tabr->isCoreIdle[slice]   = 1;
    tabr->isCoreLocked[slice] = 0;
    tabr->filp[slice] = NULL;
    spin_unlock(&tabr->reserve_lock);
    if (tabr->perf_enable) {
        record_perf_event(tabr, slice, SN_PERF_TYPE_STOP);
    }

    //signal waiting processes
    wake_up_interruptible_all(&tabr->reserve_queue);

	if (slice == 0) {
		tabr->loading[0].tv_e = ktime_get();
		tabr->loading[0].time_cnt +=
			ktime_to_us(ktime_sub(tabr->loading[0].tv_e,
				tabr->loading[0].tv_s));
	} else {
		tabr->loading[1].tv_e = ktime_get();
		tabr->loading[1].time_cnt +=
			ktime_to_us(ktime_sub(tabr->loading[1].tv_e,
				tabr->loading[1].tv_s));
	}
    return 0;
}

static int xabr_subsys_reset_keep(struct xabr_scaler *tabr, u32 slice)
{
    u32 val;
    unsigned int loop=500;
    struct sys_ctrl_cfg_regs *psub;
    struct sys_ctrl_cfg_regs abr_sys_cfg[MAX_SLICE_NUM] = {
        [0] = {"xabr_0", S1_SYS_CON_OFF, SYS_CTL_XABR,   0x24, 0x1, 0x124, 0x1, 0x71c, 0x1, 0x2, 0x52c, 0x1, 0x20},
        [1] = {"xabr_1", S2_SYS_CON_OFF, SYS_CTL_XABR,   0x24, 0x1, 0x124, 0x1, 0x71c, 0x1, 0x2, 0x52c, 0x1, 0x20}
    };

    if (slice >= MAX_SLICE_NUM) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: soft reset requested for slice_id 0x%x not supported\n", slice);
        return EINVAL;
    }
    psub = &abr_sys_cfg[slice];

    /* ---- Assert Soft Reset ----- */
    // adb_setting
    val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->adb_setting);
    writel(val&~(psub->adb_setting_clear_bits), tabr->tdev->bar2_virt + psub->slice_off + psub->adb_setting);

    //poll b5, until 0
    val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->adb_setting);
    while(val & psub->adb_setting_polling_bits) {
        --loop;
        msleep(1);
        val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->adb_setting);
        if (!loop) {
            sn_pri(tabr->tdev,SN_ERR,"%s: system controller soft reset error!\n",psub->name);
            break;
        }
    }
    if(!loop)
        return EFAULT;

    //clr b0 clk_ctrl
    val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->clock_ctrl);
    writel(val&~(psub->clock_ctrl_bits), tabr->tdev->bar2_virt + psub->slice_off + psub->clock_ctrl);

    //clr b0 sys_reset
    val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->sys_reset);
    writel(val&~(psub->sys_reset_bits), tabr->tdev->bar2_virt + psub->slice_off + psub->sys_reset);

    return 0;
}

static int xabr_subsys_reset_release(struct xabr_scaler *tabr, u32 slice)
{
    u32 val;
    unsigned int loop=500;
    struct sys_ctrl_cfg_regs *psub;
    struct sys_ctrl_cfg_regs abr_sys_cfg[MAX_SLICE_NUM] = {
        [0] = {"xabr_0", S1_SYS_CON_OFF, SYS_CTL_XABR,   0x24, 0x1, 0x124, 0x1, 0x71c, 0x1, 0x2, 0x52c, 0x1, 0x20},
        [1] = {"xabr_1", S2_SYS_CON_OFF, SYS_CTL_XABR,   0x24, 0x1, 0x124, 0x1, 0x71c, 0x1, 0x2, 0x52c, 0x1, 0x20}
    };

    if (slice >= MAX_SLICE_NUM) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: soft reset requested for slice_id %d not supported\n", slice);
        return EINVAL;
    }
    psub = &abr_sys_cfg[slice];

    /* ---- Release Soft Reset ----- */
    //set b0 clk_ctrl
    val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->clock_ctrl);
    writel(val|psub->clock_ctrl_bits, tabr->tdev->bar2_virt + psub->slice_off + psub->clock_ctrl);

    //set b0 sys_reset
    val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->sys_reset);
    writel(val|psub->sys_reset_bits, tabr->tdev->bar2_virt + psub->slice_off + psub->sys_reset);

    // adb_setting
    val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->adb_setting);
    writel(val|psub->adb_setting_clear_bits, tabr->tdev->bar2_virt + psub->slice_off + psub->adb_setting);

    //poll b5, until 1
    loop = 500;
    val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->adb_setting);
    while((val & psub->adb_setting_polling_bits) == 0x0) {
        --loop;
        msleep(1);
        val = readl(tabr->tdev->bar2_virt + psub->slice_off + psub->adb_setting);
        if (!loop) {
            sn_pri(tabr->tdev,SN_ERR,"%s: system controller soft reset error!\n",psub->name);
            break;
        }
    }
    if(!loop)
        return EFAULT;

    //clear core reservation flags
    xabr_release_core(tabr, slice);

    //Re-enable the interrupts

    //XABR_CORE_HW enable crc, wdt
    writel(0x03,  tabr->subsys[slice][XABR_CORE_HW].hwreg + XABR_SC_ADDR_CRC_WDT);
    //XABR_CORE_HW enable interrupts
    writel(0x01,  tabr->subsys[slice][XABR_CORE_HW].hwreg + XABR_SC_ADDR_GIE);
    writel(0x01,  tabr->subsys[slice][XABR_CORE_HW].hwreg + XABR_SC_ADDR_IER);

    return 0;
}

static int xabr_soft_reset(struct xabr_scaler *tabr, u32 slice)
{
    xabr_subsys_reset_keep(tabr, slice);
    xabr_subsys_reset_release(tabr, slice);
    return 0;
}

#if defined(USE_OSAL)
static void osal_handler(struct work_struct* workPtr) {
    sn_osal_work* work = (sn_osal_work*) workPtr;
    osal_command* command = work->cmdList.cmds;
    osal_response* response = work->rspList.rsps;
    int i;
    trace_printk("EXECUTING SCALER WORK\n");
    work->rspList.num_rsps = work->cmdList.num_cmds;
    for (i = 0; i < work->cmdList.num_cmds; ++i, ++command, ++response) {
        int j;
        response->header = command->header;
        for (j = 0; j < command->header.num_data_words; ++j) {
            response->data[j] = command->data[j] * 2; // do something
        }
    }
    sn_osal_finish_work(work);
}
#endif

/*
 * Calculate the utilization of abr scaler in one second.
 * It is scaler used time, from reserve a idle
 * core to release it, add up the time period in one second.
 * So the total used time divide one second is the utilization.
 */
static void xabr_loading_timer_isr(struct timer_list *t)
{
	struct xabr_scaler *tabr = from_timer(tabr, t, loading_timer);
	mod_timer(&tabr->loading_timer, jiffies + LOADING_TIME*HZ);

	tabr->loading[0].time_cnt_saved = tabr->loading[0].time_cnt;
	tabr->loading[1].time_cnt_saved = tabr->loading[1].time_cnt;

	tabr->loading[0].time_cnt = 0;
	tabr->loading[1].time_cnt = 0;

	tabr->loading[0].total_time = LOADING_TIME * 1000000;
	tabr->loading[1].total_time = LOADING_TIME * 1000000;
}

/* Display encoder utilization which is the average of the two core. */
static ssize_t xabr_util_show(struct device *dev,
				 struct device_attribute *attr,
				 char *buf)
{

	struct sn_misc_tdev *mtdev = dev_get_drvdata(dev);
	struct sn_tranx_t *tdev = mtdev->tdev;
	struct xabr_scaler *tabr = tdev->modules[SN_MODULE_XABR];
	unsigned long average_loading;

	average_loading = (tabr->loading[0].time_cnt_saved +
			tabr->loading[1].time_cnt_saved) * 100 /
			(tabr->loading[0].total_time +
			tabr->loading[1].total_time);
	if (average_loading > 100)
		average_loading = 100;

	return sprintf(buf, "%ld%%\n", average_loading);
}

static DEVICE_ATTR_RO(xabr_util);

static struct attribute *trans_xabr_sysfs_entries[] = {
	&dev_attr_xabr_util.attr,
	NULL
};

static struct attribute_group trans_xabr_attribute_group = {
	.name = NULL,
	.attrs = trans_xabr_sysfs_entries,
};

static int xabr_perf(struct sn_tranx_t* tdev, __u32 ipId, __u32 cmd, __u32 arg) {
    struct xabr_scaler* tabr = tdev->modules[SN_MODULE_XABR];
    if (cmd == SN_PERF_CMD_START) {
        tabr->perf_enable = 1;
    } else {
        tabr->perf_enable = 0;
    }
    return 0;
}

int xabr_scaler_init(struct sn_tranx_t *tdev)
{
    int i, j;
    int result = -1;
    struct xabr_scaler *tabr;
    int total_iosize;
    void *tmp_mem;
    const char *core_str[XABR_CORE_MAX] = {"core_scaler", "core_dec400"};

    tabr = (struct xabr_scaler *)kzalloc(sizeof(*tabr), GFP_KERNEL);
    if (!tabr) {
        sn_pri(tdev, SN_ERR, "xabr: kmalloc failed\n");
        return -ENOMEM;
    }

    tdev->modules[SN_MODULE_XABR] = tabr;
    tabr->tdev                    = tdev;
    total_iosize                  = init_core_array(tabr);
    if (!total_iosize) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: total core count is zero\n");
        goto err_free_tabr;
    }

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        sys_configure(tdev, i, SYS_CTL_XABR);
    }

    init_waitqueue_head(&tabr->wait_queue);
    init_waitqueue_head(&tabr->reserve_queue);
    spin_lock_init(&tabr->irq_lock);
    spin_lock_init(&tabr->reserve_lock);

    result = reserve_io(tabr);
    if(result < 0) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: reserve io failed\n");
        goto err_free_tabr;
    }

    tmp_mem = vzalloc(total_iosize);
    if (!tmp_mem) {
        sn_pri(tabr->tdev, SN_ERR, "xabr: malloc regs mem failed.\n");
        goto err_free_tabr;
    }

    tabr->regs_shadow = tmp_mem;
    for (i = 0; i < MAX_SLICE_NUM; i++) {
        for (j = 0; j < tabr->num_cores; ++j) {
            if (tabr->subsys[i][j].iosize) {
                tabr->subsys[i][j].shadow = tmp_mem;
                tmp_mem += tabr->subsys[i][j].iosize;
            }
        }
    }

    //register irq for all cores
    for (i = 0; i < MAX_SLICE_NUM; i++) {
        for (j = 0; j < tabr->num_cores; ++j) {
            if(tabr->subsys[i][j].irq > 0) {
                switch(j) {
                    case XABR_CORE_HW:
                        result = request_irq(tabr->subsys[i][j].irq, xabr_scaler_isr,
                                             IRQF_SHARED, tabr->subsys[i][j].name, (void *)tabr);
                        break;

                    case XABR_CORE_DEC400:
                        result = request_irq(tabr->subsys[i][j].irq, xabr_dec400_isr,
                                             IRQF_SHARED, tabr->subsys[i][j].name, (void *)tabr);
                        break;
                }
                if(result != 0) {
                    if(result == -EINVAL) {
                        sn_pri(tdev, SN_ERR, "xabr: Bad IRQ:%d or handler\n",
                               tabr->subsys[i][j].irq);
                    } else if(result == -EBUSY) {
                        sn_pri(tdev, SN_ERR, "xabr: IRQ:%d busy\n",
                               tabr->subsys[i][j].irq);
                    }
                    goto err_free_irq;
                } else {
                    sn_pri(tdev, SN_INF, "xabr:slice%d %s: pci_irq %d registered!\n",
                           i, core_str[j], tabr->subsys[i][j].irq);
                }
            } else {
                sn_pri(tdev, SN_INF, "xabr:slice%d %s: irq not in use!\n",
                       i,core_str[j]);
            }
        }
    }

    for (i= 0; i < MAX_SLICE_NUM; i++) {
        //XABR_CORE_HW enable crc, wdt
        writel(0x03,  tabr->subsys[i][XABR_CORE_HW].hwreg + XABR_SC_ADDR_CRC_WDT);
        //XABR_CORE_HW enable interrupts
        writel(0x01,  tabr->subsys[i][XABR_CORE_HW].hwreg + XABR_SC_ADDR_GIE);
        writel(0x01,  tabr->subsys[i][XABR_CORE_HW].hwreg + XABR_SC_ADDR_IER);

        tabr->isCoreIdle[i]   = 1;
        tabr->isCoreLocked[i] = 0;
        tabr->filp[i] = NULL;
    }

	result = sysfs_create_group(&tdev->misc_dev->this_device->kobj,
				&trans_xabr_attribute_group);
	if (result) {
		sn_pri(tdev, SN_ERR,
			 "encoder: failed to create sysfs device attributes\n");
		goto err_free_irq;
	}
	tabr->loading_timer.expires = jiffies + LOADING_TIME*HZ;
	timer_setup(&tabr->loading_timer, xabr_loading_timer_isr, 0);
	add_timer(&tabr->loading_timer);

    tabr->perf_enable = 0;
    tabr->perf_handle = sn_perf_register(tdev, "xabr", SN_PERF_ID_XABR, xabr_perf);
    sn_pri(tdev, SN_INF, "xabr: module initialization done\n");
    return 0;

err_free_irq:
    for (i= 0; i < MAX_SLICE_NUM; i++) {
        for (j = 0; j < tabr->num_cores; ++j) {
            if(tabr->subsys[i][j].irq > 0)
                free_irq(tabr->subsys[i][j].irq, (void *)tabr);
        }
    }
    vfree(tabr->regs_shadow);

err_free_tabr:
    kfree(tabr);
    tdev->modules[SN_MODULE_XABR] = NULL;
    sn_pri(tdev, SN_ERR, "xabr: module not inserted\n");
    return result;
}

int xabr_scaler_release(struct sn_tranx_t *tdev)
{
    struct xabr_scaler *tabr = tdev->modules[SN_MODULE_XABR];
    int i,j;

    del_timer_sync(&tabr->loading_timer);
    /* free the IRQ */
    for (i= 0; i < MAX_SLICE_NUM; i++) {
        for (j = 0; j < tabr->num_cores; ++j) {
            if(tabr->subsys[i][j].irq > 0)
                free_irq(tabr->subsys[i][j].irq, (void *)tabr);
        }
    }

    //release lock on all cores
    xabr_release_core(tabr, 0);
    xabr_release_core(tabr, 1);

    sysfs_remove_group(&tdev->misc_dev->this_device->kobj,
                        &trans_xabr_attribute_group);
    vfree(tabr->regs_shadow);
    kfree(tabr);
    tdev->modules[SN_MODULE_XABR] = NULL;

    sn_pri(tdev, SN_DBG, "xabr: remove module done.\n");
    return 0;
}

long xabr_scaler_ioctl(struct file *filp,
                       unsigned int cmd,
                       unsigned long argp,
                       struct sn_tranx_t *tdev)
{
    long ret = 0;
	int rv = 0;
    u32 timeout_sec;
    u32 perf_cnt[MAX_CHANNELS+1]; //0:irq status, 1..16: perf_counters
    u32 usr_args[2];
    int lock_status;
    u32 slice;
    struct core_desc core_des;
    struct xabr_scaler *tabr = tdev->modules[SN_MODULE_XABR];
    void __user *arg = (void __user *)argp;

    switch(cmd)
    {
        case XABR_IOCTL_WAIT_DONE:
        case XABR_IOCH_RESERVE:
        case XABR_IOCTL_SOFT_RESET:
        case XABR_IOCH_RELEASE:
        case PWRTEST_XABR_RESET_KEEP:
        case PWRTEST_XABR_RESET_RELEASE:
        case XABR_IOCTL_TWO_SLICE_SOFT_RESET:
            break;

        default:
            ret = copy_from_user(&core_des, (void *)arg, sizeof(struct core_desc));
            if (ret) {
                sn_pri(tdev, SN_ERR, "xabr: ioctl copy_from_user failed %d\n", ret);
                return -EFAULT;
            }
            break;
    }

    switch (cmd) {
    case XABR_IOCTL_READ_REG:
        ret = xabr_read_regs(tabr, &core_des);
        break;

    case XABR_IOCTL_WRITE_REG:
        ret = xabr_write_regs(tabr, &core_des);
        break;

    case XABR_IOCTL_WAIT_DONE:
        ret = copy_from_user(perf_cnt, (u32 *)arg, sizeof(perf_cnt));
        if (ret) {
            sn_pri(tdev, SN_ERR, "xabr: XABR_IOCTL_WAIT_DONE copy_from_user failed %d\n", ret);
            return -EFAULT;
        }
        timeout_sec  = perf_cnt[0];
        slice        = perf_cnt[1];
        ret = wait_xabr_ready(tabr, slice, timeout_sec, perf_cnt);
        rv = copy_to_user((u32 *)arg, perf_cnt, sizeof(perf_cnt));
        if (rv) {
            sn_pri(tdev, SN_ERR, "xabr: XABR_IOCTL_WAIT_DONE copy_to_user failed %d\n", rv);
        }
        break;

    case XABR_IOCH_RESERVE:
        lock_status = -1;
        ret = copy_from_user(usr_args, (u32 *)arg, sizeof(usr_args));
        if (ret) {
            sn_pri(tdev, SN_ERR, "xabr: XABR_IOCH_RESERVE copy_from_user failed %d\n", ret);
            return -EFAULT;
        }
        timeout_sec = usr_args[0];
        slice       = usr_args[1];
        ret = xabr_reserve_core(tabr, filp, slice, timeout_sec, &lock_status);
        rv = copy_to_user((u32 *)arg, &lock_status, sizeof(lock_status));
        if (rv) {
            sn_pri(tdev, SN_ERR, "xabr: XABR_IOCH_RESERVE copy_to_user failed %d\n", rv);
        }
        break;

    case XABR_IOCH_RELEASE:
        if(__get_user(slice, (u32 *)arg)) {
            sn_pri(tdev, SN_ERR, "xabr: XABR_IOCH_RELEASE __get_user failed\n");
            return -EFAULT;
        }
        ret = xabr_release_core(tabr, slice);
        break;

    case XABR_IOCTL_TWO_SLICE_SOFT_RESET:
        ret = xabr_soft_reset(tabr, 0);
        ret |= xabr_soft_reset(tabr, 1);
        break;

    case XABR_IOCTL_SOFT_RESET:
        if( __get_user(slice, (u32 *)arg)) {
            sn_pri(tdev, SN_ERR, "xabr: XABR_IOCTL_SOFT_RESET __get_user failed\n");
            return -EFAULT;
        }
        ret = xabr_soft_reset(tabr, slice);
        break;
    case PWRTEST_XABR_RESET_KEEP:
        __get_user(slice, (u32 *)arg);
        ret = xabr_subsys_reset_keep(tabr, slice);
        break;

    case PWRTEST_XABR_RESET_RELEASE:
        __get_user(slice, (u32 *)arg);
        ret = xabr_subsys_reset_release(tabr, slice);
        break;

    default:
        sn_pri(tdev, SN_ERR,
            "xabr: %s, cmd:0x%x error.\n", __func__, cmd);
        ret = -EINVAL;
    }
    return ret;
}

void xabr_close(struct sn_tranx_t *tdev, struct file *filp)
{
    struct xabr_scaler *tabr = tdev->modules[SN_MODULE_XABR];
    int i;
    for(i=0;i<MAX_SLICE_NUM;i++){
        if(tabr->filp[i] == filp) {
            xabr_release_core(tabr, i);
            sn_pri(tdev, SN_DBG, "xabr: close slice id = %d filp = %p\n", i, filp);
        }
    }
}